@@ -163,6 +163,21 @@ disclaimer.
 #ifdef CONFIG_MTD_NAND_NFC
 #define MTD_NAND_NFC
 #endif
+#ifdef CONFIG_MTD_NAND_NFC_INIT_RESET
+#define MTD_NAND_NFC_INIT_RESET
+#endif
+#ifdef CONFIG_MTD_NAND_NFC_GANG_SUPPORT
+#define MTD_NAND_NFC_GANG_SUPPORT
+#endif
+#ifdef CONFIG_MTD_NAND_NFC_MLC_SUPPORT
+#define MTD_NAND_NFC_MLC_SUPPORT
+#endif
+#ifdef CONFIG_MTD_NAND_NFC_INIT_RESET
+#define MTD_NAND_NFC_INIT_RESET
+#endif
+#ifdef CONFIG_MTD_NAND_NFC_NEGLECT_RNB
+#define MTD_NAND_NFC_NEGLECT_RNB
+#endif
 #ifdef CONFIG_MV_INCLUDE_PDMA
 #define MV_INCLUDE_PDMA
 #endif
@@ -188,12 +203,10 @@ disclaimer.
 #ifdef CONFIG_ARMADA_XP_REV_Z1
 #define MV88F78X60_Z1
 #endif
-#if defined(CONFIG_ARMADA_XP_REV_A0) || defined(CONFIG_ARMADA_XP_A0_WITH_B0)
+#ifdef CONFIG_ARMADA_XP_REV_A0
 #define MV88F78X60_A0
 #endif
-#if defined(CONFIG_ARMADA_XP_REV_B0) && !defined(CONFIG_ARMADA_XP_A0_WITH_B0)
-#define MV88F78X60_B0
-#endif
+
 /****************************************************************/
 /************* General    configuration ********************/
 /****************************************************************/
