@@ -20,7 +20,7 @@
 #define LSP_PG_SZ_VER  ""
 #endif
 
-#define LSP_VERSION    "AXP_2.0.6" LSP_PG_SZ_VER
+#define LSP_VERSION    "AXP_6.0.0" LSP_PG_SZ_VER
 
 static inline void arch_idle(void)
 {
@@ -29,7 +29,7 @@ static inline void arch_idle(void)
 
 static inline void arch_reset(char mode, const char *cmd)
 {
-	printk("Reseting...\n");
+	printk(KERN_EMERG "Reseting...\n");
 	mvBoardReset();
 	while (1);/* This should never be reached */
 }
