@@ -47,7 +47,6 @@
 #define IPC_NET_ALLOC_TIME(ptr)     (ptr + 1)
 #define IPC_NET_DATA_PTR(ptr)       (ptr + 2)
 
-static void set_multicast_list(struct net_device *dev);
 static void ipc_net_tx_timeout( struct net_device *dev);
 int ipc_net_start(struct net_device *dev);
 int ipc_net_stop( struct net_device *dev);
@@ -81,23 +80,12 @@ static const struct net_device_ops mv_ipc_netdev_ops = {
 	.ndo_open = ipc_net_start,
 	.ndo_stop = ipc_net_stop,
 	.ndo_start_xmit = ipc_net_xmit,
-	.ndo_set_multicast_list = set_multicast_list,
 	.ndo_set_mac_address = ipc_net_set_address,
 	.ndo_change_mtu = NULL,
 	.ndo_tx_timeout = ipc_net_tx_timeout,
 	.ndo_get_stats  = ipc_net_get_stats,
 };
 
-
-
-/****************************************************************************************
- * set_multicast_list()                                 				        		*
- *   fake multicast ability function 													*
- ***************************************************************************************/
-static void set_multicast_list(struct net_device *dev)
-{
-}
-
 /****************************************************************************************
  * ipc_net_tx_timeout()                                 				        		*
  *   transmit timeout function function (dummy)											*
