@@ -35,8 +35,9 @@ static MV_U32 l2fw_jhash_iv;
 static int numHashEntries;
 
 struct eth_port_l2fw **mv_eth_ports_l2fw;
+static inline int       mv_eth_l2fw_rx(struct eth_port *pp, int rx_todo, int rxq);
 static inline MV_STATUS mv_eth_l2fw_tx(struct eth_pbuf *pkt, struct eth_port *pp,
-									   int withXor, struct neta_rx_desc *rx_desc);
+					   int withXor, struct neta_rx_desc *rx_desc);
 
 
 void printBufVirtPtr(MV_BUF_INFO *pBuf)
@@ -392,7 +393,7 @@ static int mv_eth_poll_l2fw(struct napi_struct *napi, int budget)
 		if (rx_queue == -1)
 			break;
 
-		count = mv_eth_rx_l2f(pp, budget, rx_queue);
+		count = mv_eth_l2fw_rx(pp, budget, rx_queue);
 		rx_done += count;
 		budget -= count;
 		if (budget > 0)
@@ -400,7 +401,7 @@ static int mv_eth_poll_l2fw(struct napi_struct *napi, int budget)
 			 ~((1 << rx_queue) << NETA_CAUSE_RXQ_OCCUP_DESC_OFFS);
 	}
 #else
-	rx_done = mv_eth_rx_l2f(pp, budget, CONFIG_MV_ETH_RXQ_DEF);
+	rx_done = mv_eth_l2fw_rx(pp, budget, CONFIG_MV_ETH_RXQ_DEF);
 	budget -= rx_done;
 #endif /* (CONFIG_MV_ETH_RXQ > 1) */
 
@@ -671,14 +672,17 @@ static inline MV_STATUS mv_eth_l2fw_tx(struct eth_pbuf *pkt, struct eth_port *pp
 	int txq = pp->txq[smp_processor_id()];
 	read_lock(&pp->rwlock);
 	txq_ctrl = &pp->txq_ctrl[pp->txp * CONFIG_MV_ETH_TXQ + txq];
-	spin_lock(&txq_ctrl->queue_lock);
+
+	mv_eth_lock(txq_ctrl, pp);
 
 	if (txq_ctrl->txq_count >= mv_ctrl_txdone)
 		mv_eth_txq_done(pp, txq_ctrl);
 	/* Get next descriptor for tx, single buffer, so FIRST & LAST */
 	tx_desc = mv_eth_tx_desc_get(txq_ctrl, 1);
 	if (tx_desc == NULL) {
-		spin_unlock(&txq_ctrl->queue_lock);
+
+		mv_eth_unlock(txq_ctrl, pp);
+
 		read_unlock(&pp->rwlock);
 		/* No resources: Drop */
 		pp->dev->stats.tx_dropped++;
@@ -705,19 +709,24 @@ static inline MV_STATUS mv_eth_l2fw_tx(struct eth_pbuf *pkt, struct eth_port *pp
 	if (withXor) {
 		if (!xorReady()) {
 			mvOsPrintf("MV_DROPPED in %s\n", __func__);
+
+			mv_eth_unlock(txq_ctrl, pp);
+
+			read_unlock(&pp->rwlock);
 			return MV_DROPPED;
 		}
 	}
-	mvNetaTxqPendDescAdd(pp->port, pp->txp, 0, 1);
+	mvNetaTxqPendDescAdd(pp->port, pp->txp, txq, 1);
+
+	mv_eth_unlock(txq_ctrl, pp);
 
-	spin_unlock(&txq_ctrl->queue_lock);
 	read_unlock(&pp->rwlock);
 
 	return MV_OK;
 }
 
 
-inline int mv_eth_rx_l2f(struct eth_port *pp, int rx_todo, int rxq)
+static inline int mv_eth_l2fw_rx(struct eth_port *pp, int rx_todo, int rxq)
 {
 	struct eth_port  *new_pp;
 	L2FW_RULE *l2fw_rule;
