@@ -31,7 +31,6 @@ disclaimer.
 #include <linux/init.h>
 #include <linux/device.h>
 #include <linux/string.h>
-#include <linux/sysdev.h>
 #include <linux/version.h>
 #include <asm/uaccess.h>
 #include <linux/proc_fs.h>
@@ -511,7 +510,8 @@ static inline void xor_put(struct xor_channel* xch)
 static inline void dmac_clean_dcache_line(void* addr)
 {
 //#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,26)
-#if defined CONFIG_SHEEVA_ERRATA_ARM_CPU_6043 || defined(CONFIG_SHEEVA_ERRATA_ARM_CPU_6124) || defined(CONFIG_SHEEVA_ERRATA_ARM_CPU_6076)
+
+#if defined CONFIG_SHEEVA_ERRATA_ARM_CPU_6043 || defined CONFIG_SHEEVA_ERRATA_ARM_CPU_6076
 	__asm__ __volatile__ ("mcr p15, 0, %0, c7, c14, 1" : : "r" (addr));
 #else
 	__asm__ __volatile__ ("mcr p15, 0, %0, c7, c10, 1" : : "r" (addr));
