@@ -22,15 +22,11 @@
 #include <linux/rtnetlink.h>
 #include <linux/if_ether.h>
 #include <linux/slab.h>
+#include <linux/mv_nfp.h>
 #include <net/sock.h>
 
 #include "br_private.h"
 
-#if defined(CONFIG_MV_ETH_NFP_FDB_LEARN)
-int nfp_hook_del_br(int ifindex);
-int nfp_hook_del_port_from_br(int bridge_if, int port_if);
-#endif /* CONFIG_MV_ETH_NFP_FDB_LEARN */
-
 /*
  * Determine initial path cost based on speed.
  * using recommendations from 802.1d standard
@@ -173,18 +169,20 @@ void br_dev_delete(struct net_device *dev, struct list_head *head)
 
 	list_for_each_entry_safe(p, n, &br->port_list, list) {
 
-#if defined(CONFIG_MV_ETH_NFP_FDB_LEARN)
-	nfp_hook_del_port_from_br(br->dev->ifindex, p->dev->ifindex);
-#endif /* CONFIG_MV_ETH_NFP_FDB_LEARN */
+#if defined(CONFIG_MV_ETH_NFP_LEARN) || defined(CONFIG_MV_ETH_NFP_LEARN_MODULE)
+	if (nfp_mgr_p->nfp_hook_br_port_del)
+		nfp_mgr_p->nfp_hook_br_port_del(br->dev->ifindex, p->dev->ifindex);
+#endif /* CONFIG_MV_ETH_NFP_LEARN */
 
 		del_nbp(p);
 	}
 
 	del_timer_sync(&br->gc_timer);
 
-#if defined(CONFIG_MV_ETH_NFP_FDB_LEARN)
-	nfp_hook_del_br(br->dev->ifindex);
-#endif /* CONFIG_MV_ETH_NFP_FDB_LEARN */
+#if defined(CONFIG_MV_ETH_NFP_LEARN) || defined(CONFIG_MV_ETH_NFP_LEARN_MODULE)
+	if (nfp_mgr_p->nfp_hook_br_del)
+		nfp_mgr_p->nfp_hook_br_del(br->dev->ifindex);
+#endif /* CONFIG_MV_ETH_NFP_LEARN */
 
 	br_sysfs_delbr(br->dev);
 	unregister_netdevice_queue(br->dev, head);
@@ -311,10 +309,11 @@ int br_min_mtu(const struct net_bridge *br)
 /*
  * Recomputes features using slave's features
  */
-u32 br_features_recompute(struct net_bridge *br, u32 features)
+netdev_features_t br_features_recompute(struct net_bridge *br,
+	netdev_features_t features)
 {
 	struct net_bridge_port *p;
-	u32 mask;
+	netdev_features_t mask;
 
 	if (list_empty(&br->port_list))
 		return features;
@@ -440,9 +439,10 @@ int br_del_if(struct net_bridge *br, struct net_device *dev)
 	if (!p || p->br != br)
 		return -EINVAL;
 
-#if defined(CONFIG_MV_ETH_NFP_FDB_LEARN)
-	nfp_hook_del_port_from_br(br->dev->ifindex, p->dev->ifindex);
-#endif /* CONFIG_MV_ETH_NFP_FDB_LEARN */
+#if defined(CONFIG_MV_ETH_NFP_LEARN) || defined(CONFIG_MV_ETH_NFP_LEARN_MODULE)
+	if (nfp_mgr_p->nfp_hook_br_port_del)
+		nfp_mgr_p->nfp_hook_br_port_del(br->dev->ifindex, p->dev->ifindex);
+#endif /* CONFIG_MV_ETH_NFP_LEARN */
 
 	del_nbp(p);
 
