@@ -1,8 +1,15 @@
 #!/bin/bash
 
-echo " * Version: 3.0"
+echo " * Version: 3.2"
 
 # LOG:
+# 3.2:
+#   1. adding support for -l 0 to avoid network setup.
+#   2. updating smb.conf to improve single link performance.
+# 3.1:
+#   1. mount point updated.
+#   2. XOR affinity update.
+#   3. RAID-6 support.
 # 3.0:
 #   1. first all-in-one script to include: affinity, bonding, nas_init
 #   2. PAGE_SIZE automatic set
@@ -16,7 +23,7 @@ echo " * Version: 3.0"
 #   3. splice enabled for XFS.
 # 2.4:
 #   1. support RAID-1
-#   2. splice enabled by default 
+#   2. splice enabled by default
 # 2.3:
 #   1. fixing the mkdir share for loop
 # 2.2:
@@ -24,22 +31,11 @@ echo " * Version: 3.0"
 # 2.1:
 #   1. setting coal to 100 in net queue 0 and fixing for in NETQ
 
-
-#
-# t= topology [sd|rd0|rd1|rd5]
-# p= prepare drives (fdisk)
-# m= mkfs+mdadm
-# b= burn new rootfs to /dev/sda1
-# f= file system format (ext4/xfs/btrfs)
-# s= for SSD medium
-# l= for largepage support (64k)
-# h= for HDDs number (4/8) in RAID5 configuration
-
 PREPARE_HDD="no"
 ZAP_ROOTFS="no"
 MKFS="no"
-TOPOLOGY="sd"
-HDD_NUM="1"
+TOPOLOGY="rd5"
+HDD_NUM="4"
 ARCH=`uname -m`
 FS="ext4"
 NFS="no"
@@ -105,53 +101,53 @@ while getopts "l:pbmzn:sf:t:h:" flag; do
 		*)			do_error "Usage: file-system: ext4|xfs|btrfs" ;;
 	    esac
 	    ;;
-	
+
 	s)	SSD="yes"
 	    ;;
-	
+
 	b)	ZAP_ROOTFS="yes"
 	    ROOTFS_TARBALL="$OPTARG"
 	    ;;
-	
+
 	m)	MKFS="yes"
 	    ;;
 
 	z)	SYSDISKEXIST="yes"
 	    ;;
-	
+
 	n)	PARTNUM="$OPTARG"
 	    ;;
-	
+
 	p)	PREPARE_HDD="yes"
 	    ;;
-	
+
 	l)	LINK_NUM=$OPTARG
 	    case "$OPTARG" in
-		1|2|4) ;;
-		*)	do_error "Usage: links options: 1|2|4" ;;
+		0|1|2|4) ;;
+		*)	do_error "Usage: links options: 0|1|2|4" ;;
 	    esac
 	    ;;
-	
+
 	t)	TOPOLOGY=$OPTARG
 	    case "$OPTARG" in
-	sd|rd0|rd1|rd5) ;;
-		*)	do_error "Usage: drive toplogy: sd|rd0|rd1|rd5" ;;
+	sd|rd0|rd1|rd5|rd6) ;;
+		*)	do_error "Usage: drive toplogy: sd|rd0|rd1|rd5|rd6" ;;
 	    esac
 	    ;;
-	
+
 	h)	HDD_NUM=$OPTARG
 	    case "$OPTARG" in
 		4|8) ;;
 		*)	do_error "Usage: drive toplogy: 4|8" ;;
 	    esac
 	    ;;
-	
+
 	*)	echo "Usage: $0"
 	    echo "           -f <ext4|xfs|btrfs|fat32>: file system type ext4, xfs, btrfs or fat32"
-	    echo "           -t <sd|rd0|rd1|rd5>: drive topology"
+	    echo "           -t <sd|rd0|rd1|rd5|rd6>: drive topology"
 	    echo "           -n <num>: partition number to be mounted"
-	    echo "           -m: mkfs/mdadm"
-	    echo "           -p: prepare drives"
+	    echo "           -m: create RAID and FD (mkfs/mdadm)"
+	    echo "           -p: prepare drives (fdisk)"
 	    echo "           -h <num>: number of HDDs to use"
 	    echo "           -l <num>: number of links to use"
 	    echo "           -b <rootfs_tarball_path>:  path to rootfs tarball to be placed on /dev/sda2"
@@ -167,6 +163,7 @@ case "$TOPOLOGY" in
     rd0)	echo -ne "RAID0\n" ;;
     rd1)	echo -ne "RAID1\n" ;;
     rd5)	echo -ne "RAID5\n" ;;
+    rd6)	echo -ne "RAID6\n" ;;
     *)	do_error "Invalid Topology" ;;
 esac
 
@@ -188,7 +185,7 @@ if [ "$MKFS" == "yes" ]; then
     esac
 fi
 echo -ne "******************************************\n"
-[[ "$TOPOLOGY" == "rd0" || "$TOPOLOGY" == "rd1" || "$TOPOLOGY" == "rd5" ]] && [ ! -e "$(which mdadm)" ] && do_error "missing mdadm in rootfs (aptitude install mdadm)"
+[[ "$TOPOLOGY" == "rd0" || "$TOPOLOGY" == "rd1" || "$TOPOLOGY" == "rd5" || "$TOPOLOGY" == "rd6" ]] && [ ! -e "$(which mdadm)" ] && do_error "missing mdadm in rootfs (aptitude install mdadm)"
 
 #if [[ -e "$(which smbd)" && -e "$(which nmbd)" ]]; then
 #    echo -n "* Starting Samba daemons"
@@ -218,9 +215,9 @@ fi
 
 sleep 2
 
-if [ `mount | grep $MNT_DIR | grep -v grep | wc -l` != 0 ]; then
-    do_error "Unable to unmount $MNT_DIR"
-fi
+#if [ `mount | grep mnt | grep -v grep | wc -l` != 0 ]; then
+#    do_error "Unable to unmount /mnt"
+#fi
 echo -ne "[Done]\n"
 
 #[ -e /dev/md_d0 ] && mdadm --manage --stop /dev/md_d0
@@ -263,10 +260,26 @@ elif [ "$TOPOLOGY" == "rd5" ]; then
 	fi
 	PARTSIZE="20GB"
     fi
+elif [ "$TOPOLOGY" == "rd6" ]; then
+    if [ "$HDD_NUM" == "8" ]; then
+	if [ "$SYSDISKEXIST" == "yes" ]; then
+	    DRIVES="b c d e f g h i"
+	else
+	    DRIVES="a b c d e f g h"
+	fi
+	PARTSIZE="10GB"
+    elif [ "$HDD_NUM" == "4" ]; then
+	if [ "$SYSDISKEXIST" == "yes" ]; then
+	    DRIVES="b c d e"
+	else
+	    DRIVES="a b c d"
+	fi
+	PARTSIZE="20GB"
+    fi
 fi
 
-# Checking drives existence 
-#for drive in `echo $DRIVES`; do 
+# Checking drives existence
+#for drive in `echo $DRIVES`; do
  #   if [ `fdisk -l |grep "Disk /dev/sd$drive" |grep -v grep |wc -l` == 0 ]; then
 #	do_error "Drives assigned (/dev/sd$drive) is not persent!"
  #   fi
@@ -275,7 +288,7 @@ fi
 echo -ne " * Preparing disks partitions: "
 if [ "$PREPARE_HDD" == "yes" ]; then
     [ ! -e "$(which fdisk)" ] && do_error "missing fdisk in rootfs"
-    
+
     set -o verbose
     if [ "$SSD" == "no" ]; then
 	for drive in `echo $DRIVES`; do echo -e "c\no\nn\np\n1\n4096\n+${PARTSIZE}\nt\n83\nw\n" | fdisk -u /dev/sd${drive}; done
@@ -285,14 +298,14 @@ if [ "$PREPARE_HDD" == "yes" ]; then
     fi
     set +o verbose
     fdisk -ul
-    
+
     blockdev --rereadpt /dev/sd${drive} || do_error_reboot "The partition table has been altered, please reboot device and run nas_init again"
 fi
 echo -ne "[Done]\n"
 
 if [ "$ZAP_ROOTFS" == "yes" ]; then
     [ ! -e "$ROOTFS_TARBALL" ] && do_error "missing rootfs tarball, use option -b <rootfs_tarball_path>"
-    
+
     mkfs.ext3 ${ROOTFS_DEVICE}
     mount ${ROOTFS_DEVICE} $MNT_DIR
     cd $MNT_DIR
@@ -306,6 +319,42 @@ if [ "$ZAP_ROOTFS" == "yes" ]; then
     umount $MNT_DIR
     sleep 2
 fi
+echo -ne " * Network setup:              "
+if [ "$LINK_NUM" == "2" ]; then
+    set -o verbose
+    ifconfig eth0 0.0.0.0 down
+    ifconfig eth1 0.0.0.0 down
+
+    ifconfig bond0 192.168.0.5 netmask 255.255.255.0 up
+    ifenslave bond0 eth0 eth1
+
+    ifconfig bond0 down
+    echo balance-alb > /sys/class/net/bond0/bonding/mode
+    ifconfig bond0 up
+    set +o verbose
+elif [ "$LINK_NUM" == "4" ]; then
+    set -o verbose
+    ifconfig eth0 0.0.0.0 down
+    ifconfig eth1 0.0.0.0 down
+    ifconfig eth2 0.0.0.0 down
+    ifconfig eth3 0.0.0.0 down
+
+    ifconfig bond0 192.168.0.5 netmask 255.255.255.0 up
+    ifenslave bond0 eth0 eth1 eth2 eth3
+
+    ifconfig bond0 down
+    echo balance-alb > /sys/class/net/bond0/bonding/mode
+    ifconfig bond0 up
+    set +o verbose
+elif [ "$LINK_NUM" == "1" ]; then
+    set -o verbose
+    ifconfig eth0 192.168.0.5 netmask 255.255.255.0 up
+    set +o verbose
+elif [ "$LINK_NUM" == "0" ]; then
+    echo -ne "Skipping network setup\n"
+fi
+echo -ne "[Done]\n"
+
 echo -ne " * Network optimization:       "
 
 if [ "$ARCH" == "armv5tel" ]; then
@@ -333,7 +382,7 @@ elif [[ "$ARCH" == "armv6l" || "$ARCH" == "armv7l" ]]; then
 	for (( i=0; i<=$NETQ; i++ )); do
 	    echo "$j $i 100" > /sys/devices/platform/neta/gbe/rxq_time_coal
 	done
-	    
+
 	ethtool -k eth$j > /dev/null
     done
     set +o verbose
@@ -350,7 +399,7 @@ if [ "$TOPOLOGY" == "rd5" ]; then
     set -o verbose
 
     for drive in `echo $DRIVES`; do echo -e 1024 > /sys/block/sd${drive}/queue/read_ahead_kb; done
-    
+
     if [ "$MKFS" == "yes" ]; then
 	[ -e /dev/md0 ] && mdadm --manage --stop /dev/md0
 
@@ -418,6 +467,81 @@ if [ "$TOPOLOGY" == "rd5" ]; then
     set +o verbose
     echo -ne "[Done]\n"
 
+elif [ "$TOPOLOGY" == "rd6" ]; then
+    echo -ne " * Starting RAID6 build:       "
+    for drive in `echo $DRIVES`; do PARTITIONS="${PARTITIONS} /dev/sd${drive}${PARTNUM}"; done
+
+    set -o verbose
+
+    for drive in `echo $DRIVES`; do echo -e 1024 > /sys/block/sd${drive}/queue/read_ahead_kb; done
+
+    if [ "$MKFS" == "yes" ]; then
+	[ -e /dev/md0 ] && mdadm --manage --stop /dev/md0
+
+	for partition in `echo $DRIVES`; do mdadm --zero-superblock /dev/sd${partition}1; done
+
+	if [ "$SSD" == "no" ]; then
+	    echo "y" | mdadm --create -c 128 /dev/md0 --level=6 -n $HDD_NUM --force $PARTITIONS
+	else
+			# most SSD use eraseblock of 512, so for performance reasons we use it
+	    echo "y" | mdadm --create -c 512 /dev/md0 --level=6 -n $HDD_NUM --force $PARTITIONS
+	fi
+
+	sleep 2
+
+	if [ `cat /proc/mdstat  |grep md0 |wc -l` == 0 ]; then
+	    do_error "Unable to create RAID device"
+	fi
+
+	if [ "$FS" == "ext4" ]; then
+	    if [ "$SSD" == "no" ]; then
+		if [ "$HDD_NUM" == "8" ]; then
+		    mkfs.ext4 -j -m 0 -T largefile -b 4096 -E stride=32,stripe-width=224 -F /dev/md0
+		elif [ "$HDD_NUM" == "4" ]; then
+		    mkfs.ext4 -j -m 0 -T largefile -b 4096 -E stride=32,stripe-width=96 -F /dev/md0
+		fi
+	    else
+		mkfs.ext4 -j -m 0 -T largefile -b 4096 -E stride=128,stripe-width=384 -F /dev/md0
+	    fi
+	elif [ "$FS" == "xfs" ]; then
+	    mkfs.xfs -f /dev/md0
+	elif [ "$FS" == "btrfs" ]; then
+	    mkfs.btrfs /dev/md0
+	fi
+    else
+		# need to reassemble the raid
+	mdadm --assemble /dev/md0 --force $PARTITIONS
+
+	if [ `cat /proc/mdstat  |grep md0 |wc -l` == 0 ]; then
+	    do_error "Unable to assemble RAID device"
+	fi
+    fi
+
+    if [ "$FS" == "ext4" ]; then
+	mount -t ext4 /dev/md0 $MNT_DIR -o noatime,data=writeback,barrier=0,nobh
+    elif [ "$FS" == "xfs" ]; then
+	mount -t xfs /dev/md0 $MNT_DIR -o noatime,nodirspread
+    elif [ "$FS" == "btrfs" ]; then
+	mount -t btrfs /dev/md0 $MNT_DIR -o noatime
+    fi
+
+    if [ `mount | grep $MNT_DIR | grep -v grep | wc -l` == 0 ]; then
+	do_error "Unable to mount FS"
+    fi
+
+    if [ "$LARGE_PAGE" == "65536" ]; then
+	echo 256 > /sys/block/md0/md/stripe_cache_size
+    else
+	echo 4096 > /sys/block/md0/md/stripe_cache_size
+    fi
+    /bin/echo -e 4096 > /sys/block/md0/queue/read_ahead_kb
+
+    for drive in `echo $DRIVES`; do echo noop > /sys/block/sd${drive}/queue/scheduler; done
+    echo 4 > /proc/sys/vm/dirty_ratio
+    echo 2 > /proc/sys/vm/dirty_background_ratio
+    set +o verbose
+    echo -ne "[Done]\n"
+
 elif [ "$TOPOLOGY" == "rd1" ]; then
     echo -ne " * Starting RAID1 build:       "
     for drive in `echo $DRIVES`; do PARTITIONS="${PARTITIONS} /dev/sd${drive}${PARTNUM}"; done
@@ -494,13 +618,13 @@ elif [ "$TOPOLOGY" == "rd0" ]; then
 	else
 	    echo "y" | mdadm --create -c 256 /dev/md0 --level=0 -n 2 --force $PARTITIONS
 	fi
-	    
+
 	sleep 2
 
 	if [ `cat /proc/mdstat  |grep md0 |wc -l` == 0 ]; then
 	    do_error "Unable to create RAID device"
 	fi
-	  
+
 	echo 256 > /sys/block/md0/md/stripe_cache_size
 	if [ "$FS" == "ext4" ]; then
 	    if [ "$SSD" == "no" ]; then
@@ -513,7 +637,7 @@ elif [ "$TOPOLOGY" == "rd0" ]; then
 	fi
     else
 	mdadm --assemble /dev/md0 --force $PARTITIONS
-	
+
 	if [ `cat /proc/mdstat  |grep md0 |wc -l` == 0 ]; then
 	    do_error "Unable to create RAID device"
 	fi
@@ -539,7 +663,7 @@ elif [ "$TOPOLOGY" == "sd" ]; then
     PARTITIONS="/dev/sd${DRIVES}${PARTNUM}"
     echo -ne " * Starting single disk:       "
     set -o verbose
-    
+
     echo -e 1024 > /sys/block/sd${DRIVES}/queue/read_ahead_kb
 
     if [ "$FS" == "fat32" ]; then
@@ -553,7 +677,7 @@ elif [ "$TOPOLOGY" == "sd" ]; then
     elif [ "$FS" == "ext4" ]; then
 	if [ "$MKFS" == "yes" ]; then
 	    [ -e /dev/md0 ] && mdadm --manage --stop /dev/md0
-	    
+
 	    mkfs.ext4 -j -m 0 -b 4096 -O large_file,extents -F $PARTITIONS
 	fi
 	mount -t ext4 $PARTITIONS $MNT_DIR -o noatime,data=writeback,barrier=0
@@ -562,7 +686,7 @@ elif [ "$TOPOLOGY" == "sd" ]; then
 	    [ -e /dev/md0 ] && mdadm --manage --stop /dev/md0
 	    mkfs.btrfs -m single -d single $PARTITIONS
 	fi
-	
+
 	mount -t btrfs $PARTITIONS $MNT_DIR -o noatime
     else
 	if [ "$MKFS" == "yes" ]; then
@@ -610,7 +734,7 @@ if [ "$FS" != "NONE" ]; then
 	echo ' dns proxy = no'				>>  /etc/smb.conf
 	echo ' max log size = 200'			>>  /etc/smb.conf
 	echo ' log level = 0'				>>  /etc/smb.conf
-	echo ' socket options = IPTOS_LOWDELAY TCP_NODELAY SO_SNDBUF=131072 SO_RCVBUF=131072' >>  /etc/smb.conf
+	echo ' socket options = IPTOS_LOWDELAY TCP_NODELAY' >>  /etc/smb.conf
 	echo ' local master = no'			>>  /etc/smb.conf
 	echo ' dns proxy = no'				>>  /etc/smb.conf
 	echo ' ldap ssl = no'				>>  /etc/smb.conf
@@ -624,14 +748,13 @@ if [ "$FS" != "NONE" ]; then
 	echo ' disable netbios = yes'			>>  /etc/smb.conf
 	echo ' csc policy = disable'			>>  /etc/smb.conf
 	if [ "$FS" == "btrfs" ]; then
-			# crash identified with these FS
+	# crash identified with these FS
 	    echo '# min receivefile size = 128k'	>>  /etc/smb.conf
 	    echo '# strict allocate = yes'		>>  /etc/smb.conf
 	else
 	    echo ' min receivefile size = 128k'	>>  /etc/smb.conf
 	    echo ' strict allocate = yes'		>>  /etc/smb.conf
 	fi
-	echo ' allocation roundup size = 10485760'	>>  /etc/smb.conf
 	echo ''						>>  /etc/smb.conf
 	echo '[public]'					>>  /etc/smb.conf
 	echo ' comment = my public share'		>>  /etc/smb.conf
@@ -665,48 +788,52 @@ echo -ne " * Setting up affinity:        "
 if [[ "$ARCH" == "armv6l" || "$ARCH" == "armv7l" ]]; then
     if [ "$CPU_COUNT" = "4" ]; then
 	set -o verbose
-	
-		# XOR Engines
+
+	# XOR Engines
 	echo 1 > /proc/irq/51/smp_affinity
-	echo 1 > /proc/irq/52/smp_affinity
-	
-		# ETH
+	echo 2 > /proc/irq/52/smp_affinity
+	echo 4 > /proc/irq/94/smp_affinity
+	echo 8 > /proc/irq/95/smp_affinity
+
+	# ETH
 	echo 1 > /proc/irq/8/smp_affinity
 	echo 2 > /proc/irq/10/smp_affinity
 	echo 4 > /proc/irq/12/smp_affinity
 	echo 8 > /proc/irq/14/smp_affinity
-	
-		# SATA
+
+	# SATA
 	echo 2 > /proc/irq/55/smp_affinity
-	
-		# PCI-E SATA controller
+
+	# PCI-E SATA controller
 	echo 4 > /proc/irq/99/smp_affinity
 	echo 8 > /proc/irq/103/smp_affinity
-	
+
 	set +o verbose
     elif [ "$CPU_COUNT" == "2" ]; then
 	set -o verbose
-	
-		# XOR Engines
+
+	# XOR Engines
 	echo 1 > /proc/irq/51/smp_affinity
-	echo 1 > /proc/irq/52/smp_affinity
-	
-		# ETH
+	echo 2 > /proc/irq/52/smp_affinity
+	echo 1 > /proc/irq/94/smp_affinity
+	echo 2 > /proc/irq/95/smp_affinity
+
+	# ETH
 	echo 1 > /proc/irq/8/smp_affinity
 	echo 2 > /proc/irq/10/smp_affinity
-	
-		# SATA
+
+	# SATA
 	echo 2 > /proc/irq/55/smp_affinity
-	
-		# PCI-E SATA controller
+
+	# PCI-E SATA controller
 	echo 2  > /proc/irq/99/smp_affinity
-	
+
 	set +o verbose
     fi
 fi
 echo -ne "[Done]\n"
 case "$TOPOLOGY" in
-    rd1 | rd5)
+    rd1 | rd5 | rd6)
 		#watch "cat /proc/mdstat|grep finish"
 	;;
 esac
